/*  TV Fixer for GP2X

	File: settv.c
	
  Copyright (C) 2006  Kounch
	Parts Copyright (c) 2005 Rlyeh

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>     //Entrada/Salida basica
#include <unistd.h>    //Para poder salir al menu de GP2X con execl, obtener pid
#include <sys/mman.h>  //Para poder hacer mmap y acceder a los registros de memoria
#include <fcntl.h>     //Para poder tener acceso al dispositivo /dev/mem
#include <stdlib.h>    //Para convertir de cadena a long, leer y escribir variables de entorno, etc.

//Si compilamos para GP2X
#ifdef __GP2X__

#endif

#include "settv.h" //Cabecera principal del programa


/*Variables globales*/
char textoError[MAXLEN]; //Para almacenar la descripcion del ultimo error producido
unsigned long tiempo; //El tiempo que hay que esperar (en decimas de segundo) entre iteraciones

unsigned long   gp2x_dev[8]={0,0,0,0,0,0,0,0};
volatile unsigned short *gp2x_memregs;
static volatile unsigned long  *gp2x_memregl;


/******************************************************************/
/*Rutina Main, aqui analizamos los argumentos de linea de comandos*/
/******************************************************************/

int main (int argc, char **argv) 
{
  char cad[MAXLEN];
  int horizontal,vertical;
  int bpp,window,fin;
  int x1,x2,y1,y2;
  float width,height,tvwidth,tvheight;
  float escalaw,escalah;
  unsigned long tick;
  int i,corregir;
  
  /*Resoluciones consideradas como correctas
    320.00 x 240.00 Estandar (SDL de GamePark y Minimal Lib sin escalado por HW)
    159.67 x 143.65 GNUBoy2X HW Scaling
    320.00 x 223.36
    383.48 x 268.03
    223.15 x 200.58
    239.51 x 167.30
    255.87 x 191.82
    303.64 x 211.97
    351.41 x 263.65
    239.51 x 143.65
    255.87 x 197.96
    207.44 x 160.29
    335.71 x 251.39
    223.15 x 177.81 MAME HW Scaling Normal
    223.15 x 223.36
    223.15 x 268.03
    287.28 x 200.58
    223.15 x 167.30
    239.51 x 191.82
    223.75 x 212.25
    239.51 x 263.65
    191.74 x 143.65
    247.36 x 197.96
    247.36 x 160.29
    239.51 x 251.39
    255.87 x 177.81 MAME HW Scaling Rotado
    255.94 x 223.88 SquidgeSNES HW Scaling
  */
    
  int resx[27]={320,159,320,383,223,239,255,303,351,239,255,207,335,223,223,223,287,223,239,223,239,191,247,247,239,255,255};
  int resy[27]={240,143,223,268,200,167,191,211,263,143,197,160,251,177,223,268,200,167,191,212,263,143,197,160,251,177,223};


  printf("\nSetTV "__VERSION_PROG__" para GP2X. Copyright 2006 Kounch\n");
  
  /*Comprobacion de argumentos*/
  if ((argc != 2) && (argc!=4)) {
    sprintf (textoError, "Error: Nmero de argumentos incorrecto, eran %i, debera ser uno o tres.\n", --argc);
    muestraError();
    return(1); 
  }
  
  /*Se toma el primer argumento como dato de tiempo*/
  tiempo=atoi(argv[1]);
  printf("Tiempo de espera entre iteraciones: %i\n",tiempo);
    
  //Obtiene el pid de un archivo ./loop.pid, mata la tarea con ese pid y luego borra el archivo
  getLoopPid();
  
  //Guarda  nuestro nuevo pid en un archivo ./loop.pid
  if (tiempo)
    setLoopPid();

#ifdef __GP2X__
  /*Inicializacin para acceder a la memoria de la consola */
  if(!gp2x_dev[2])  gp2x_dev[2] = open("/dev/mem", O_RDWR);
  gp2x_memregl=(unsigned long  *)mmap(0, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, gp2x_dev[2], 0xc0000000);
  gp2x_memregs=(unsigned short *)gp2x_memregl;

  fin=0;
  
	do
  {
    corregir=1;
    muestraEstado();
 
    horizontal=gp2x_memregs[0x2906>>1];	
    vertical=gp2x_memregs[0x2908>>1];	
    printf ("Pantalla: %i x %i\n",horizontal,vertical);
    
    bpp=(gp2x_memregs[0x28DA>>1]>>9)&0x3;

    /*Comprobando TV*/
    if(gp2x_memregs[0x2800>>1]&0x100)
    {
      printf ("TV-Out activado\n");
      
      /*Coordenadas de ventana 0 RGB (comunes a PAL y NTSC)*/
      x1=0;     
      x2=669;
      y1=0;
      
      escalaw=489.0; //Factor de escala RGB Horizontal para TV (comun a PAL y NTSC)
      if (gp2x_memregs[0x2818>>1]  == 287) //Comprobacion si la altura de la pantalla es para PAL
      {
        printf ("PAL detectado\n");
        escalah=274.0; //Factor de escala RGB Vertical para TV PAL
        y2=279; //Coordenadas de ventana 0 RGB para PAL
        
      }
      else if (gp2x_memregs[0x2818>>1]  == 239) //Comprobacion si la altura de la pantalla es para NTSC
      {
        printf ("NTSC detectado\n");
        escalah=331.0; //Factor de escala RGB Vertical para TV NTSC
        y2=231; //Coordenadas de ventana 0 RGB para NTSC
      }
    }
    else //Suponemos que esta en modo LCD (TV-Out desactivado)
    {
      printf ("TV-Out desactivado\n");
      escalaw=1024.0; //Escala RGB Horizontal para LCD
      escalah=320.0; //Escala RGB Vertical para LCD
      x1=0;
      x2=319;
      y1=0;
      y2=239;
      
      //Como no estamos en salida de Tv, no corregimos nada
      corregir=0;
    }

    //Calculamos el tamao de pantalla si fuera para LCD
    width=(horizontal/1024.0)*320.0;
    height=(vertical/(320.0*bpp))*240.0;
    printf("Guess LCD: %.2f x %.2f\n",width,height);
    //Calculamos el tamao de pantalla si fuera para TV
    tvwidth=(horizontal/escalaw)*320.0;
    tvheight=(vertical/(escalah*bpp))*240.0;
    printf("Guess  TV: %.2f x %.2f\n",tvwidth,tvheight);
    printf("Guess iTV: %i x %i\n",(int)tvwidth,(int)tvheight); 
    
    
    //Tests especiales para el GNBUBoy2X con soporte de TV
    if ( (width==160.0) && (height==240.0) )
      width=320.0;
    if ( (width==80.0) && (height==144.0) )
      width=160.0;
    if ( (((int)width)==95) && (height==144.0) )
      width=191.88;    
    

    //Comparamos la resolucion tipo TV 
    for (i=0;i<27;i++)
    {
      printf ("Comprobacion %i; %ix%i\n",i,resx[i],resy[i]);
	    if (((int)tvwidth)==resx[i])
	    {
		    if (((int)tvheight)==resy[i])
		    {
			    //Resolucion valida
          printf ("Dimensiones validas\n");
			    //deinit();
			    //exit(0);
          corregir=0;
          break;
		    }
		    else if ((int)tvheight==(2*resy[i]) || (resy[i]==(2*(int)tvheight)))
		    {
			    //Escala vertical equivocada
          printf ("Escala vertical incorrecta\n");
			    width=tvwidth;
			    height=resy[i];
          corregir=1;
          break;
		    }
	    }
    }
    
    //Si se pedia una resolucion por linea de comandos, usamos esa
    if (argc==4)
    {
      width=atof(argv[2]);
      height=atof(argv[3]);
      corregir=1;
    }
    
    //Si hemos determinado que hay que hacer ajustes, los aplicamos ahora
    if (corregir)
    {
      printf ("width: %.2f x height: %.2f\n", width, height);
    
      // Ajuste de escala horizontal
      gp2x_memregs[0x2906>>1]=(unsigned short)((float)escalaw *(width/320.0));
      // Ajuste de escala vertical
      gp2x_memregl[0x2908>>2]=(unsigned long)((float)escalah *bpp *(height/240.0)); 
    
      // Ventana 0 RGB
      window=0;
      gp2x_memregs[(0x28e2+window*8)>>1]=x1;     
      gp2x_memregs[(0x28e4+window*8)>>1]=x2;
      gp2x_memregs[(0x28e6+window*8)>>1]=y1;
      gp2x_memregs[(0x28e8+window*8)>>1]=y2;  
        
      muestraEstado();
    }


    if (tiempo)
    {
      //Esto carga demasiado la CPU
      //tick=gp2x_memregl[0x0A00>>2]/7372;
      //while (gp2x_memregl[0x0A00>>2]/7372-tick < (tiempo*100)); //Esperar tiempo (en decimas de segundo)
      
      //Esto no carga la CPU de forma apreciable
      sleep(tiempo);
    }
    else
      fin=1;
  }
  while (!fin);

  deinit();
  
  return(0);
}


/*Funcion que cierra todos los dispositivos que se hayan abierto (basada en Minimal Lib)*/
int deinit(void)
{
	{ int i; for(i=0;i<8;i++) if(gp2x_dev[i]) close(gp2x_dev[i]); }
	return (0);
}
  

/*Funcion que muestra valores interesantes de la configuracion de la pantalla*/
int muestraEstado(void)
{
  int horizontal,vertical,bpp,width;
  int window,x1,x2,y1,y2;
  
  /*Escala horizontal*/
  horizontal=gp2x_memregs[0x2816>>1];	
  /*Escala vertical*/
  vertical=gp2x_memregs[0x2818>>1];	
  printf ("Pantalla: %i x %i\n",horizontal,vertical);
  
  bpp=(gp2x_memregs[0x28DA>>1]>>9)&0x3;
  printf ("BPP: %i\n",bpp);
  
  /*Escala horizontal*/
  horizontal=gp2x_memregs[0x2906>>1];	
  /*Escala vertical*/
  vertical=gp2x_memregs[0x2908>>1];	
  printf ("Escala: %i x %i\n",horizontal,vertical);
  /*RGB Width*/
  width=gp2x_memregs[0x290C>>1];	
  printf ("Anchura RGB: %i\n",width);
  /*Coordenadas de ventanas*/
  for(window=0;window<4;window++) //windows 0..3
  {
    x1=gp2x_memregs[(0x28e2+window*8)>>1];     
    x2=gp2x_memregs[(0x28e4+window*8)>>1];
    y1=gp2x_memregs[(0x28e6+window*8)>>1];
    y2=gp2x_memregs[(0x28e8+window*8)>>1];
    
    printf ("Ventana %i; %i,%i,%i,%i\n",window,x1,x2,y1,y2);
  }
  
  return(0);
}

#endif


//Obtiene el pid de un archivo ./loop.pid, mata la tarea con ese pid y luego borra el archivo
int getLoopPid(void)
{
  FILE *fp;
  char str[MAXLEN];
  int pid;
  
  fp = fopen("./loop.pid", "rt");
  
  if(fp == NULL) return (0);
  
  fgets(str,MAXLEN,fp);
  pid=atoi(str);
  
  fclose(fp);
  
  sprintf (str,"kill -9 %i;rm -f ./loop.pid",pid);
  system(str);
  sync();
  
  return (pid);
}


//Guarda nuestro pid en un archivo ./loop.pid
int setLoopPid(void)
{
  FILE *fp;
  pid_t pid;
  
  /*Obtenemos nuestro pid*/
  pid=getpid();
  /*Lo escribimos en un archivo*/
  fp = fopen("./loop.pid", "w");
  fprintf(fp,"%i",pid);  
  fclose(fp);
  sync();
  
  return (0);
}


int muestraError(void)
  {
  int n,m;
  unsigned short col;
  char cad[MAXLEN];
  
  fprintf (stderr, textoError);
  
  return (0);
  }
