/*  TV Fixer for GP2X

	File: tvfixer.c
	
  Copyright (C) 2006  Kounch
	Parts Copyright (C) 2005 Hermes/PS2Reality and modified by Vimacs and god_at_hell(LCD-part)
	Parts Copyright (c) 2005 Rlyeh

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <stdio.h>     //Entrada/Salida basica
#include <string.h>    //Manejo de cadenas de texto
#include <stdlib.h>    //Utilidades basicas
#include <iconv.h>     //Para convertir de UTF-8 a ISO-8859-1
#include <sys/stat.h>  //Acceso a funciones del sistema
#include <dirent.h>    //Manejo de directorios
#include <unistd.h>    //Necesario por pthread, entre otras cosas


//Si compilamos para GP2X
#ifdef __GP2X__
  #include "minimal.h" //Libreria principal para GP2X
  #include "cpuctrl.h" //Para ajustar la velocidad de la CPU

  #define CPU_SPEED 66 //Velocidad de CPU que usa el programa por defecto
  
  /*Colores por defecto*/
  unsigned COLOR_TITULO=0x00ff00; //color del titulo con el copyright, etc
  unsigned COLOR_COMANDO=0xcccccc; //color de la linea de comandos
  unsigned COLOR_LISTADO=0x00ffff; //color del listado de opciones
  unsigned COLOR_SELECCION=0xffffff; //color del archivo seleccionado en el listado
  unsigned COLOR_AYUDA=0xff0000; //color del texto de ayuda "Pulsa Start para Salir"
  unsigned COLOR_ERROR=0x00ffff; //color del texto en los mensajes de error
  unsigned COLORFONDO=0xBB8888; // color de fondo
  
  char firmVersion [10]; //para guardar la version de firmware detectada
  unsigned short cpuver; //para guardar la version de firmware de forma numerica
  unsigned dclock; //para guardar la velocidad de reloj de la pantalla
  int lfreq; //para guardar la frecuencia de reloj original

  //extern unsigned short splash[]; //Imagen de inicio
  extern unsigned short fondo[]; //Imagen de fondo
#endif

#include "tvfixer.h" //Cabecera principal del programa
#include "lang.h" //Funciones y cadenas de texto en distintos idiomas


/*Variables globales que se generan desde el archivo de configuracion*/
int tiempo; //Valor de tiempo a pasar al programa que ajusta la salida de TV (settv)
int activo; //Para indicar si esta activo el loop o no
char lenguaje[4]; //Para saber el idioma activo
char tv[4]; //Para saber si esta activa la salida a TV o el LCD
char tipoTV[5]; //Para saber el estandar de TV activo

char *langStrings [LANGSTRINGSN]; //Vector que almacena todos los mensajes de texto del programa
char textoError [MAXLEN]; //para guardar mensajes de error para la funcion muestraError

/******************************************************************/
/*Rutina Main, aqui analizamos los argumentos de linea de comandos*/
/******************************************************************/

int main (int argc, char **argv) 
{
  char cad[MAXLEN];
  int loopPid; 

  /*Inicializacion del lenguaje por defecto*/
  initLang(DEFAULTLANG);
  strcpy (lenguaje,DEFAULTLANG);
  
#ifdef __GP2X__
  /*Inicializamos la consola */
  gp2x_init(1000, 16, 44100,16,1,50,0);
 
  //Borramos la pantalla
  //ClearScreen(splash);
  ClearScreen(fondo);

  gp2x_video_RGB_flip(0); //Actualizamos la pantalla

  //Obtenemos el estado de la CPU, etc.
  gp2x_getcpu();
  //Cambiamos a CPU_SPEED MHz
  set_speed_clock(CPU_SPEED);
#endif

  /*Si hay problemas al leer el archivo de configuracion, nos salimos*/
  if (leeArchivo("./tvfixer.conf"))
  {
    sprintf (textoError, langStrings[2]);
    muestraError();
    return(1);
  }
  
  /*Comprobamos si hay un archivo con el pid del loop*/
  if (loopPid=getLoopPid())
  {
    activo=1;
    printf (langStrings[7],loopPid);
  }
  else
  {
    activo=0;
    printf (langStrings[8]);
  }

#ifdef __GP2X__
  //Esperamos un par de segundos
  gp2x_timer_delay(2000);
#endif
  
  /*Si todo ha ido bien, llamamos a la rutina principal*/
  if ( interfaz() )
  {
    sprintf (textoError, langStrings[3]);
    muestraError();
#ifdef __GP2X__
    //Recuperamos el estado de los registros de memoria 
    gp2x_setcpu();
    
    //Salimos al menu de la consola
    gp2x_deinit();
#endif
    return(2);
  }
  
#ifdef __GP2X__
  //Recuperamos el estado de los registros de memoria
  gp2x_setcpu();
  
  if (activo)
  {
    //Lanzamos el bucle con el refresco indicado
    sprintf(cad,"./settv %i >/dev/null 2>&1 &",tiempo);
    system(cad);
  }
  
  //Guardamos los cambios en el archivo de configuracion
  saveConf();

   //Salimos al men de la consola
  gp2x_deinit();
  #endif

  return(0);
  }


/*Recorre el archivo llamando a las funciones que cargan las variables globales del programa*/
int leeArchivo(char *nombre)
{
  FILE *fp;
  char str[MAXLEN];
  int line;
  
  printf(langStrings[4],nombre);

  fp = fopen(nombre, "rt");
  if(fp == NULL)
  {
    //Error al intentar abrir el archivo, seguramente no existe
    //printf (langStrings[5]);
    return (1);
  }

  for (line=1;fgets(str,MAXLEN,fp);line++)
    {
	if (analizaVar(str))
	  {
	  /*Se ha cargado una variable*/
	  }
	else
	  {
	  /*Es un comentario*/
	  }
	}

  fclose(fp);
  
  return (0);
}


/*Toma una linea de texto e intenta separarla en base a encontrar un signo '=', obteniendo una clave y un valor*/
/*Si lo consigue, compara con las variables globales a cargar y, si ha encontrado un valor, se lo asigna*/
int analizaVar (char *linea)
{
  char *clave;
  char *valor;
  char charHex[3];
  char valorHex[3];
  int i;
  
  if ( (linea[0] != '#') && (linea[0] != '\n') )
  {
    clave = strtok (linea, "=");
    valor = strtok (NULL, "\n");
    
    //Carga de idioma (si lo hay)
    if ( !strcmp (clave, "langCode") && (valor != NULL) ) 
    {
      initLang(valor);
      strcpy(lenguaje,valor);
    }
 
    //Carga el valor por defecto para el tiempo (si lo hay)
    if ( !strcmp (clave, "Time") && (valor != NULL) )
      tiempo=atoi(valor);
    else
      tiempo=10;

    //Habia una variable
    return (1);
	}
  //Habia un comentario o una variable desconocida o una linea en blanco
  return (0);
}


//Obtiene el pid de un archivo ./loop.pid, mata la tarea con ese pid y luego borra el archivo
int getLoopPid(void)
{
  FILE *fp;
  char str[MAXLEN];
  int pid;
  
  fp = fopen("./loop.pid", "rt");
  
  if(fp == NULL) return (0);
  
  fgets(str,MAXLEN,fp);
  pid=atoi(str);
  
  fclose(fp);
  
  sprintf (str,"kill -9 %i;rm -f ./loop.pid",pid);
  system(str);
  sync();
    
  return (pid);
}



/* Actualiza la interfaz de usuario al recibir respuesta de la funcion leeDatos */
int interfaz(void)
{
  int x,fin;

#ifdef __GP2X__
  /*Inicializamos la pantalla */
  ClearScreen(fondo);
	gp2x_video_RGB_flip(0);
#endif
  
  x=0;
  actualizaInterfaz(x);  
  fin=0;
  do {
	switch (leeDatos())
	  {
    case 1:
      //Pulsado Select
      break;
    case 2:
      //Pulsada Y
      break;
    case 3:
      //Pulsada A
      break;
    case 4:
      //Pulsada la X
    case 5:
      //Pulsada la B
      activo=1-activo;
      break;
	  case 6: //Orden de salida de la app
      fin=1;
      break;
    case 7:
      if (tiempo > 1) --tiempo;
      break;
    case 8:
      if (tiempo < 120) ++tiempo;
      break;

    default:
        break;
	  }	  
	if (!fin)
	  {
	  actualizaInterfaz(x);
	  }	  
	} while (!fin);
  return (0);
}


/* Funciones para GP2X */
#ifdef __GP2X__

int muestraError(void)
  {
  int n,m;
  unsigned short col;
  char cad[MAXLEN];
  int ancho=48;
  
  col=gp2x_video_RGB_color16(0xff,0,0);

  for (n=16;n<304;n++)
    {
    for(m=96;m<136;m++)
      {
      gp2x_video_RGB[0].screen16[n+(m*320)]=col;
      }
    }

  sprintf(cad,"*** ERROR ***");
  *(cad+ancho)=0; //Recortamos las lineas para que quede bien en pantalla
  v_putcad(20,13,COLOR_ERROR,cad);

  sprintf(cad,textoError);
  *(cad+ancho)=0; //Recortamos las lineas para que quede bien en pantalla
  v_putcad(3,15,COLOR_ERROR,cad);
  
  gp2x_video_RGB_flip(0);

  fprintf (stderr, textoError);
  
  leeDatos();
  
  return (0);
  }


/* Lee la entrada de los controles de la GP2X*/
int leeDatos(void)
  {
  unsigned teclasGP2X;
  unsigned long tick;

  tick=gp2x_timer_read();
  while (gp2x_timer_read()-tick < 100); //Esperar varios msec para que la interfaz no sea muy rpida


  teclasGP2X=0;
  while (1)
    {
	teclasGP2X=gp2x_joystick_read();
	
	if((teclasGP2X & GP2X_R) &&	(teclasGP2X & GP2X_L))
	  {
	  return (9);
	  }
	  
	if((teclasGP2X & GP2X_R))
	  {
	  return (8);
	  }

	if((teclasGP2X & GP2X_L))
	  {
	  return (7);
	  }

	if((teclasGP2X & GP2X_START))
	  {
	  return (6);
	  }

	if((teclasGP2X & GP2X_B))
	  {
	  return (5);
	  }

	if((teclasGP2X & GP2X_X))
	  {
	  return (4);
	  }

    if((teclasGP2X & GP2X_A))
	  {
	  return (3);
	  }

    if((teclasGP2X & GP2X_Y))
	  {
	  return (2);
	  }

    if((teclasGP2X & GP2X_SELECT))
	  {
	  return (1);
	  }
	}

  return (0);
}


/*Funcion que guarda la configuracion en un archivo ./tvfixer.conf*/
int saveConf(void)
{
    FILE *fp;
    pid_t pid;
    
    fp = fopen("./tvfixer.conf", "w");
    fprintf(fp,"langCode=%s\n",lenguaje);
    fprintf(fp,"Time=%i",tiempo);
    fclose(fp);
    sync();
    
    return (0);
}


/*Funcion que actualiza la interfaz de usuario*/
int actualizaInterfaz(int coordenada)
  {
  char cad[MAXLEN];
  unsigned colorLetra;
  int minline;
  int maxline;
  int i;
  int fila;
  char *puntero1, *puntero2;
  int ancho=50;

  /*Borramos todo*/
  ClearScreen(fondo);
  
  sprintf(cad,"** GP2X TV-Out Fixer v"__VERSION_PROG__" ** (c) 2006 Kounch");
  *(cad+ancho)=0; //Recortamos las lineas para que quede bien en pantalla
	v_putcad(3,2,COLOR_TITULO,cad);
  
  sprintf(cad,langStrings[10],tv);
  *(cad+ancho-4)=0; //Recortamos las lineas para que quede bien en pantalla
	v_putcad(9,6,COLOR_LISTADO,cad);

  sprintf(cad,langStrings[11],tipoTV);
  *(cad+ancho-4)=0; //Recortamos las lineas para que quede bien en pantalla
	v_putcad(9,8,COLOR_LISTADO,cad);
  
  if (activo)
    sprintf(cad,langStrings[12],"ON");
  else
    sprintf(cad,langStrings[12],"OFF");    
  
  *(cad+ancho-4)=0; //Recortamos las lineas para que quede bien en pantalla
	v_putcad(9,10,COLOR_LISTADO,cad);

  sprintf(cad,langStrings[13],tiempo);
  *(cad+ancho-4)=0; //Recortamos las lineas para que quede bien en pantalla
	v_putcad(9,12,COLOR_LISTADO,cad);
  
  sprintf(cad,langStrings[14]);
  *(cad+ancho)=0; //Recortamos las lineas para que quede bien en pantalla
  v_putcad(3,16,COLOR_AYUDA,cad);

  sprintf(cad,langStrings[15]);
  *(cad+ancho)=0; //Recortamos las lineas para que quede bien en pantalla
  v_putcad(3,18,COLOR_AYUDA,cad);

  sprintf(cad,langStrings[16]);
  *(cad+ancho)=0; //Recortamos las lineas para que quede bien en pantalla
  v_putcad(3,20,COLOR_AYUDA,cad);
  
  sprintf(cad,langStrings[9],firmVersion,lfreq);
  *(cad+ancho)=0; //Recortamos las lineas para que quede bien en pantalla
  v_putcad(1,26,COLOR_COMANDO,cad);

  gp2x_video_RGB_flip(0);

  return (0);
}


/*Para ajustar el reloj del 920*/
void set_speed_clock(int gp2x_speed)
{
set_display_clock_div(16+16*(gp2x_speed>=200));
set_FCLK(gp2x_speed);set_DCLK_Div(0);set_920_Div(0);
set_display_clock_div(4);
}


/* Para obtener la version de firmware, y el estado de la CPU*/
int gp2x_getcpu(void)
  {
  FILE *fp;
  char str[MAXLEN];
  unsigned sysfreq;
  int n;
  unsigned frecuencias[25]={60,66,80,100,120,133,150,166,200,210,220,240,250,266,270,275,280,285,290,295,300,305,310,315,320}; // por encima de 220 MHz solo funciona con div 1
  extern unsigned short *gp2x_memregs;
  volatile unsigned short *MEM_REG;
  
  MEM_REG=&gp2x_memregs[0];
  
  strcpy(tv,"LCD");
  strcpy(tipoTV,"---");

  if(MEM_REG[0x2800>>1]&0x100)
  {
    strcpy(tv,"TV");
    if (MEM_REG[0x2818>>1]  == 287)
      strcpy(tipoTV,"PAL");
    else if (MEM_REG[0x2818>>1]  == 239)
      strcpy(tipoTV,"NTSC");
  }
  
  fp = fopen("/usr/gp2x/version", "rt");
  if(fp == NULL)
  {
	sprintf (textoError, langStrings[17]);
	muestraError();
    return (1);
  }
  fgets(str,MAXLEN,fp);
  fclose(fp);
  str[5]=0;
  
  strcpy(firmVersion,str);
  
  if (!strcmp(firmVersion,"1.0"))
    cpuver=100;
  else
	cpuver=101;

  cpuctrl_init(); 
  Disable_940(); // desconecta el 940t
  save_system_regs();
  dclock=get_display_clock_div();
  
  sysfreq=get_freq_920_CLK();
  sysfreq*=get_920_Div()+1;
  lfreq=(sysfreq/1000000)+4;
  
  for(n=24;n>1;n--)
	{
	if(frecuencias[n]<lfreq) break; 
	}
  lfreq=frecuencias[n];
  }
  
/* Para dejar todo como estaba*/
int gp2x_setcpu(void)
  {
  //Borrado de pantalla
  ClearScreen(0);
  gp2x_video_RGB_flip(0);
  ClearScreen(0);
  gp2x_video_RGB_flip(0);

  //Recuperamos registros del sistema
  load_system_regs();
  set_display_clock_div(dclock);

  //Cerramos lo relativo a la minimal lib
  //Ojo! esta version de gp2x_deinit() debe estar modificada para que no llame al menu de GP2X
  //gp2x_deinit();
  }
  

/* se usa para 'borrar' la pantalla virtual con una imagen (no hace video_flip)*/
void ClearScreen(unsigned short *imagen)
{
int n;
unsigned short col;
unsigned short *colo;

col=0;
colo=imagen;

for(n=0;n<320*240;n++)
	{
    
  if (colo)
    gp2x_video_RGB[0].screen16[n]=*colo++;
  else
    gp2x_video_RGB[0].screen16[n]=col;
	}
}


/*Dibuja una cadena de texto con un color definido*/
void v_putcad(int x,int y,unsigned color,char *cad) 
{  
  //while(cad[0]!=0) {v_putchar(x,y,color,cad[0]);cad++;x++;}
  gp2x_fontcolor_set(NULL, (int) color);
  gp2x_printf(NULL,x*6,y*8,cad);
  }


/*Requerido por las Minimal Lib*/
void gp2x_sound_frame(void *blah, void *buff, int samples) { }

#endif
